part of masamune.firebase;

/// Firebase / Firestore Authentication class
/// Basically, anonymous authentication is performed, followed by a flow linking a specific authentication method
class FirestoreAuth extends Auth {
  /// Create a Completer that matches the class.
  /// Do not use from external class
  @override
  @protected
  Completer createCompleter() => Completer<FirestoreAuth>();
  /// Process to create a new instance
  /// Do not use from outside the class
  @override
  @protected
  T createInstance<T extends IClonable>() => FirestoreAuth._( ) as T;
  Firebase get _app {
    if( this.__app == null ) this.__app = Firebase._find(this.protocol);
    return this.__app;
  }
  Firebase __app;
  FirebaseAuth get _auth {
    if( this.__auth == null ) this.__auth = FirebaseAuth.fromApp( this._app._app );
    return this.__auth;
  }
  FirebaseAuth __auth;
  FirebaseUser _link;
  GoogleSignIn get _google {
    if( this.__google == null ) this.__google = GoogleSignIn();
    return this.__google;
  }
  GoogleSignIn __google;
  TwitterLogin get _twitter {
    if( this.__twitter == null ) this.__twitter = TwitterLogin(
      consumerKey: _twitterAPIKey,
      consumerSecret: _twitterAPISecret
    );
    return this.__twitter;
  }
  TwitterLogin __twitter;
  FacebookLogin get _facebook {
    if( this.__facebook == null ) this.__facebook = FacebookLogin();
    return this.__facebook;
  }
  FacebookLogin __facebook;
  static FirestoreAuth _find( String protocol ) {
    if( isEmpty(protocol) ) protocol = "firestore";
    String path = Texts.format( _systemPath, [protocol] );
    FirestoreAuth unit = PathMap.get<FirestoreAuth>( path );
    if( unit != null ) return unit;
    return PathMap.get<FirestoreAuth>( Texts.format( _systemPath, [Const.empty] ) );
  }

  /// Set options for authentication.
  /// 
  /// [twitterAPIKey]: Twitter API Key
  /// [twitterAPISecret]: Twitter API Secret
  static void options({String twitterAPIKey, String twitterAPISecret}){
    _twitterAPIKey = twitterAPIKey;
    _twitterAPISecret = twitterAPISecret;
  }
  static String _twitterAPIKey;
  static String _twitterAPISecret;

  /// Check if you are logged in
  /// True if logged in
  /// 
  /// [protocol]: Protocol specification
  static Future<bool> isLoggedIn({String protocol}){
    if( isEmpty(protocol) ) protocol = "firestore";
    String path = Texts.format( _systemPath, [protocol] );
    FirestoreAuth unit = PathMap.get<FirestoreAuth>( path );
    if( unit != null ) return unit._isLoggedInProcess();
    unit = FirestoreAuth._( path );
    return unit._isLoggedInProcess();
  }
  Future<bool> _isLoggedInProcess() async {
    if( this._app == null ) this.__app = await Firebase.initialize();
    if( this._auth == null ) return false;
    try {
      FirebaseUser user = await this._auth.currentUser();
      if( user != null ){
        await user.reload();
        return true;
      }
    } catch(e){ }
    return false;
  }
  
  
  /// Process sign-in
  /// Basically, anonymous authentication is performed, followed by a flow linking a specific authentication method
  /// 
  /// [protorol]: Protocol specification
  static Future<FirestoreAuth> signIn({String protocol}) => signInAnonymously(protocol:protocol);
  /// Process sign-in
  /// Perform an anonymous login
  /// 
  /// [protorol]: Protocol specification
  static Future<FirestoreAuth> signInAnonymously({String protocol}) {
    if( isEmpty(protocol) ) protocol = "firestore";
    String path = Texts.format( _systemPath, [protocol] );
    FirestoreAuth unit = PathMap.get<FirestoreAuth>( path );
    if( unit == null ){
      unit = FirestoreAuth._( path );
    } else {
      unit.init();
    }
    unit._anonymousProcess();
    return unit.future;
  }
  static const String _systemPath = "system://auth/%s";
  FirestoreAuth._([String path]) : super( path:path );

  /// Sign out
  /// 
  /// [protorol]: Protocol specification
  static Future<FirestoreAuth> signOut({String protocol}){
    if( isEmpty(protocol) ) protocol = "firestore";
    String path = Texts.format( _systemPath, [protocol] );
    FirestoreAuth unit = PathMap.get<FirestoreAuth>( path );
    if( unit == null ){
      unit = FirestoreAuth._( path );
    } else {
      unit.init();
    }
    if( isEmpty(unit.uid) || unit._link == null ){
      Log.error( "Not logged in yet. Please wait until login is successful." );
      return unit.future;
    }
    unit._signOutProcess();
    return unit.future;
  }
  Future _signOutProcess() async {
    if( this._app == null ) this.__app = await Firebase.initialize();
    if( this._auth == null ){
      this.error( "Firebase auth is not found." );
      return;
    }
    this._google?.currentUser?.clearAuthCache();
    this._twitter?.logOut();
    this._facebook?.logOut();
    await this._auth.signOut();
    this.done();
  }

  /// Link to Facebook
  /// 
  /// [protorol]: Protocol specification
  static Future<FirestoreAuth> signInToFacebook({String protocol}){
    if( isEmpty(protocol) ) protocol = "firestore";
    String path = Texts.format( _systemPath, [protocol] );
    FirestoreAuth unit = PathMap.get<FirestoreAuth>( path );
    if( unit == null ){
      unit = FirestoreAuth._( path );
    } else {
      unit.init();
    }
    unit._linkToFacebookProcess( );
    return unit.future;
  }
  Future _linkToFacebookProcess( ) async {
    try {
      if( isEmpty( this.uid ) ) await this._anonymousProcessInternal( );
      if( this._link.providerData.any((t) => t.providerId?.contains("facebook") ) ){
        this.error("This user is already linked to a Facebook account.");
        return;
      }
      FacebookLoginResult result = await this._facebook.logIn(["email"]);
      switch (result.status) {
        case FacebookLoginStatus.cancelledByUser:
          this.error("Login canceled");
          return;
        case FacebookLoginStatus.error:
          this.error(Texts.format( "Login terminated with error: %s", [result.errorMessage]));
          return;
        default: break;
      }
      AuthCredential credential = FacebookAuthProvider.getCredential(
        accessToken: result.accessToken.token
      );
      this._link = (await _link.linkWithCredential(credential)).user;
      if( this._link == null || isEmpty( this._link.uid )  ){
        this.error( "User is not found." );
        this.dispose();
        return;
      }
      Log.msg("Linked Facebook account to user: %s", [this._link.uid]);
      this.authorized( this._link.uid );
    } catch( e ) {
      this.error(e);
      this.dispose();
      return;
    }
  }

  /// Link to Twitter
  /// 
  /// [protorol]: Protocol specification
  static Future<FirestoreAuth> signInToTwitter({String protocol}){
    if( isEmpty(protocol) ) protocol = "firestore";
    String path = Texts.format( _systemPath, [protocol] );
    FirestoreAuth unit = PathMap.get<FirestoreAuth>( path );
    if( unit == null ){
      unit = FirestoreAuth._( path );
    } else {
      unit.init();
    }
    unit._linkToTwitterProcess( _twitterAPIKey, _twitterAPISecret);
    return unit.future;
  }
  Future _linkToTwitterProcess(String apiKey, String apiSecret) async {
    try {
      if( isEmpty( this.uid ) ) await this._anonymousProcessInternal( );
      if( this._link.providerData.any((t) => t.providerId?.contains("twitter") ) ){
        this.error("This user is already linked to a Twitter account.");
        return;
      }
      TwitterLoginResult result = await this._twitter.authorize();
      switch (result.status) {
        case TwitterLoginStatus.cancelledByUser:
          this.error("Login canceled");
          return;
        case TwitterLoginStatus.error:
          this.error(Texts.format( "Login terminated with error: %s", [result.errorMessage]));
          return;
        default: break;
      }
      AuthCredential credential = TwitterAuthProvider.getCredential(
        authToken: result.session.token,
        authTokenSecret: result.session.secret
      );
      this._link = (await _link.linkWithCredential(credential)).user;
      if( this._link == null || isEmpty( this._link.uid )  ){
        this.error( "User is not found." );
        this.dispose();
        return;
      }
      Log.msg("Linked Twitter account to user: %s", [this._link.uid]);
      this.authorized( this._link.uid );
    } catch( e ) {
      this.error(e);
      this.dispose();
      return;
    }
  }

  /// Link Google
  /// Google login dialog appears
  /// 
  /// [protorol]: Protocol specification
  static Future<FirestoreAuth> signInToGoogle({String protocol}) {
    if( isEmpty(protocol) ) protocol = "firestore";
    String path = Texts.format( _systemPath, [protocol] );
    FirestoreAuth unit = PathMap.get<FirestoreAuth>( path );
    if( unit == null ){
      unit = FirestoreAuth._( path );
    } else {
      unit.init();
    }
    unit._linkToGoogleProcess();
    return unit.future;
  }
  Future _linkToGoogleProcess() async {
    try {
      if( isEmpty( this.uid ) ) await this._anonymousProcessInternal( );
      GoogleSignInAccount googleCurrentUser = this._google.currentUser;
      if (googleCurrentUser == null) googleCurrentUser = await this._google.signInSilently();
      if (googleCurrentUser == null) googleCurrentUser = await this._google.signIn();
      if (googleCurrentUser == null) {
        this.error("Google user could not get.");
        return;
      }
      if( this._link.providerData.any((t) => t.providerId?.contains("google") ) ){
        this.error("This user is already linked to a Google account.");
        return;
      }
      GoogleSignInAuthentication googleAuth = await googleCurrentUser.authentication;
      AuthCredential credential = GoogleAuthProvider.getCredential(
        accessToken: googleAuth.accessToken,
        idToken: googleAuth.idToken,
      );
      this._link = (await _link.linkWithCredential(credential)).user;
      if( this._link == null || isEmpty( this._link.uid )  ){
        this.error( "User is not found." );
        this.dispose();
        return;
      }
      Log.msg("Linked Google account to user: %s", [this._link.uid]);
      this.authorized( this._link.uid );
    } catch( e ) {
      this.error(e);
      this.dispose();
      return;
    }
  }  
  Future _anonymousProcess( ) async {
    try {
      await this._anonymousProcessInternal( );
      Log.msg("Anonymous login successful: %s", [this._link.uid]);
      this.authorized( this._link.uid );
    } catch( e ) {
      this.error(e);
      this.dispose();
      return;
    }
  }
  Future _anonymousProcessInternal( ) async {
    if( this._app == null ) this.__app = await Firebase.initialize();
    if( this._auth == null ){
      this.error( "Firebase auth is not found." );
      return;
    }
    try {
      try {
        FirebaseUser user = await this._auth.currentUser();
        if( user != null ){
          await user.reload();
          this._link = user;
          Log.msg( "Logged-in user was found: %s", [this._link.uid]);
          return;
        }
      } catch(e) { }
      this._link = ( await this._auth.signInAnonymously() )?.user;
      if( this._link == null || isEmpty( this._link.uid )  ){
        this.error( "User is not found." );
        this.dispose();
        return;
      }
    } catch( e ) {
      this.error(e);
      this.dispose();
      return;
    }
  }

  /// Execute when authentication is completed
  /// Do not use from external class
  /// 
  /// [uid]: User / Universal ID
  @override
  @protected
  void authorized(String uid){
    PathTag.set(Const.uid, uid);
    PathTag.set("UserID", uid);
    super.authorized(uid);
  }
}